from django.conf import settings
from django.core.signals import setting_changed
from django.utils.module_loading import import_string

DEFAULTS = {
    "BLACKLIST_URLS": [],
    "SENSITIVE_FIELDS": ["password"],
    "REQUEST_ID_HEADER": "X-Request-ID",
    "APP_VERSION": "dev",
}


IMPORT_STRINGS = []


class LibSettings:
    """
    Taken from the djangorestframework library. Thanks.
    """

    def __init__(self, defaults=None, import_strings=None):
        self.defaults = defaults or DEFAULTS
        self.import_strings = import_strings or IMPORT_STRINGS
        self._cached_attrs = set()

    @property
    def user_settings(self):
        if not hasattr(self, "_user_settings"):
            self._user_settings = getattr(settings, "ZUPIT_LOGGING", {})
        return self._user_settings

    def __getattr__(self, attr):
        if attr not in self.defaults:
            raise AttributeError("Invalid API setting: '%s'" % attr)

        try:
            # Check if present in user settings
            val = self.user_settings[attr]
        except KeyError:
            # Fall back to defaults
            val = self.defaults[attr]

        # Coerce import strings into classes
        if attr in self.import_strings:
            val = import_string(val)

        # Cache the result
        self._cached_attrs.add(attr)

        setattr(self, attr, val)
        return val

    def reload(self):
        for attr in self._cached_attrs:
            delattr(self, attr)
        self._cached_attrs.clear()
        if hasattr(self, "_user_settings"):
            delattr(self, "_user_settings")


lib_settings = LibSettings(DEFAULTS, IMPORT_STRINGS)


def reload_lib_settings(*args, **kwargs):
    setting = kwargs["setting"]
    if setting == "ZUPIT_LOGGING":
        lib_settings.reload()


setting_changed.connect(reload_lib_settings)



import dataclasses
from datetime import datetime
from unittest import mock
from unittest.mock import patch

from django.contrib.auth.models import AnonymousUser, User
from django.test import RequestFactory, TestCase

from src.logging import ZupitLoggingMiddleware
from src.settings import lib_settings


@dataclasses.dataclass
class MockResponse:
    status_code: int


class ZupitLoggingMiddlewareTestCase(TestCase):
    def setUp(self) -> None:
        self.factory = RequestFactory()
        self.user = User.objects.create_user(username="jacob", email="jacob@…", password="top_secret")
        self.anon_user = AnonymousUser()

    @patch("logging.Logger.info")
    def test_log_userful_request_info(self, info_logger_mock):
        request = self.factory.get("/")
        request.user = self.user
        time = datetime(1901, 12, 21)

        response = MockResponse(status_code=200)

        middleware = ZupitLoggingMiddleware(get_response=mock.MagicMock(return_value=response))

        with patch("django_zupit_logging.logging.datetime", wraps=datetime) as mock_datetime:
            mock_datetime.utcnow.return_value = time
            middleware(request)

        info_logger_mock.assert_called_once_with(
            msg={
                "date_time": str(time),
                "remote_address": "127.0.0.1",
                "request_method": request.method,
                "request_path": request.path,
                "user_id": self.user.id,
                "request_id": "",
                "app_version": lib_settings.APP_VERSION,
                "params": "{}",
                "status_code": response.status_code,
            }
        )

    @patch("logging.Logger.info")
    def test_log_logged_user_request(self, info_logger_mock):
        request = self.factory.get("/")
        request.user = self.user

        middleware = ZupitLoggingMiddleware(get_response=mock.MagicMock(return_value=MockResponse(status_code=200)))
        middleware(request)

        _, kwargs = info_logger_mock.call_args
        msg = kwargs.get("msg")
        user_id = msg.get("user_id")

        info_logger_mock.assert_called_once()
        self.assertEquals(user_id, self.user.id)

    @patch("logging.Logger.info")
    def test_log_anonymous_user_request(self, info_logger_mock):
        request = self.factory.get("/")
        request.user = self.anon_user

        middleware = ZupitLoggingMiddleware(get_response=mock.MagicMock(return_value=MockResponse(status_code=200)))
        middleware(request)

        _, kwargs = info_logger_mock.call_args
        msg = kwargs.get("msg")
        user_id = msg.get("user_id")

        info_logger_mock.assert_called_once()
        self.assertEquals(user_id, "")

    @patch("logging.Logger.info")
    @patch("logging.Logger.debug")
    def test_log_successfull_request(self, info_logger_mock, debug_logger_mock):
        request = self.factory.get("/")
        request.user = self.user

        middleware = ZupitLoggingMiddleware(get_response=mock.MagicMock(return_value=MockResponse(status_code=200)))
        middleware(request)

        info_logger_mock.assert_called_once()
        debug_logger_mock.assert_called_once()

    @patch("logging.Logger.info")
    @patch("logging.Logger.warning")
    def test_log_client_error_request(self, info_logger_mock, warning_logger_mock):
        request = self.factory.get("/")
        request.user = self.user

        middleware = ZupitLoggingMiddleware(get_response=mock.MagicMock(return_value=MockResponse(status_code=400)))
        middleware(request)

        info_logger_mock.assert_called_once()
        warning_logger_mock.assert_called_once()

    @patch("logging.Logger.info")
    @patch("logging.Logger.error")
    def test_log_server_error_request(self, info_logger_mock, error_logger_mock):
        request = self.factory.get("/")
        request.user = self.user

        middleware = ZupitLoggingMiddleware(get_response=mock.MagicMock(return_value=MockResponse(status_code=500)))
        middleware(request)

        info_logger_mock.assert_called_once()
        error_logger_mock.assert_called_once()



from pathlib import Path

BASE_DIR = Path(__file__).resolve().parent.parent

SECRET_KEY = "fake-key"


INSTALLED_APPS = [
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "src",
    "tests",
]

DEFAULT_AUTO_FIELD = "django.db.models.BigAutoField"

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": BASE_DIR / "db.sqlite3",
    }
}

ROOT_URLCONF = "tests.urls"

USE_TZ = True

TIME_ZONE = "Europe/Rome"

ZUPIT_LOGGING = {
    "BLACKLIST_URLS": ["/blacklisted/"],
    "SENSITIVE_FIELDS": ["password", "credit_card_number"],
    "REQUEST_ID_HEADER": "X-Request-ID",
    "APP_VERSION": "1.2.3",
}



import json
import logging
from datetime import datetime

from src.settings import lib_settings

logger = logging.getLogger(__name__)


def clean_sensitive_info(dictionary):
    sensitive_fields = lib_settings.SENSITIVE_FIELDS
    clean_dict = {k: "***" if k in sensitive_fields else v for (k, v) in dictionary.items()}
    return clean_dict


def parse_clean_json(myjson):
    if not myjson:
        return myjson

    try:
        return clean_sensitive_info(json.loads(myjson))
    except ValueError as ex:
        logger.error("Could not parse JSON", exc_info=ex)
        return None


def get_client_ip(request):
    x_forwarded_for = request.META.get("HTTP_X_FORWARDED_FOR")
    if x_forwarded_for:
        ip = x_forwarded_for.split(",")[0]
        # Maybe it's the last, cfr. https://stackoverflow.com/questions/4581789/how-do-i-get-user-ip-address-in-django
    else:
        ip = request.META.get("REMOTE_ADDR")
    return ip


class ZupitLoggingMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        url_is_not_blacklisted = all(url not in str(request.get_full_path()) for url in lib_settings.BLACKLIST_URLS)

        if url_is_not_blacklisted:
            log_data = {
                "date_time": str(datetime.utcnow()),
                "remote_address": get_client_ip(request),
                "request_method": request.method,
                "request_path": request.path,
                "user_id": request.user.id if request.user and request.user.id else "",
                "request_id": request.headers.get(lib_settings.REQUEST_ID_HEADER, ""),
                "app_version": lib_settings.APP_VERSION,
                "params": json.dumps(request.GET),
            }

            logger.info(msg=log_data)

        response = self.get_response(request)

        if url_is_not_blacklisted:
            response_log_message = log_data

            response_log_message["status_code"] = response.status_code

            if 200 <= response.status_code < 400:
                logger.debug(response_log_message)
            elif 400 <= response.status_code < 500:
                logger.warning(response_log_message)
            elif response.status_code >= 500:
                logger.error(response_log_message)

        return response
